﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace DeoVR
{
    public class RemoteControlClient
    {
        const double ConnectionTimeout = 3.0;
        const double PingDelay = 1.0;

        public event Action OnConnected;
        public event Action<Exception> OnDisconnected;
        public event Action<RemoteApiData> OnDataReceived;

        string hostname;
        int port;

        bool lastEmptyResponse;
        Thread thread;
        List<byte> receivedData = new List<byte>();
        public Queue<RemoteApiData> sendQueue = new Queue<RemoteApiData>();

        public bool IsConnected { get { return thread != null; } }

        public void Connect(string hostname, int port)
        {
            if (thread != null)
                throw new Exception("Client is already connected!");

            if (String.IsNullOrEmpty(hostname))
                throw new Exception("Empty host name!");
            if (port <= 0 || port > 65535)
                throw new Exception("Incorrect port!");

            this.hostname = hostname;
            this.port = port;
            thread = new Thread(threadProc);
            thread.Start();
        }

        public void Send(RemoteApiData newData)
        {
            lock (sendQueue)
                sendQueue.Enqueue(newData);
        }

        public void Disconnect()
        {
            if (thread == null)
                return;

            thread.Abort();
            thread = null;
        }

        void threadProc()
        {
            receivedData.Clear();
            sendQueue.Clear();
            try
            {
                using (var client = new TcpClient(hostname, port))
                using (var stream = client.GetStream())
                {
                    OnConnected?.Invoke();
                    DateTime lastSendTime = DateTime.UtcNow;
                    DateTime lastReceiveTime = DateTime.UtcNow;
                    var buffer = new byte[1024];
                    while (client.Connected)
                    {
                        if ((DateTime.UtcNow - lastReceiveTime).Seconds >= ConnectionTimeout)
                            throw new Exception("Connection timeout! (>=" + ConnectionTimeout.ToString() + ")");

                        // send data to DeoVR
                        while (sendQueue.Count > 0)
                        {
                            RemoteApiData data = null;
                            lock (sendQueue)
                                data = sendQueue.Dequeue();
                            if (data != null)
                            {
                                sendRemoteData(data, stream);
                                lastSendTime = DateTime.UtcNow;
                            }
                        }

                        // ping DeoVR
                        if ((DateTime.UtcNow - lastSendTime).Seconds >= PingDelay)
                        {
                            sendRemoteData(null, stream);
                            lastSendTime = DateTime.UtcNow;
                        }

                        // read incoming data
                        if (stream.DataAvailable)
                        {
                            int size = stream.Read(buffer, 0, buffer.Length);
                            for (int i = 0; i < size; i++)
                                receivedData.Add(buffer[i]);

                            if (parseBuffer())
                                lastReceiveTime = DateTime.UtcNow;
                        }
                        Thread.Sleep(10);
                    }
                }
                thread = null;
                OnDisconnected?.Invoke(null);
            }
            catch (ThreadAbortException)
            {
                Thread.ResetAbort();
                thread = null;
                OnDisconnected?.Invoke(null);
            }
            catch (Exception ex)
            {
                thread = null;
                OnDisconnected?.Invoke(ex);
            }
        }

        void sendRemoteData(RemoteApiData data, NetworkStream stream)
        {
            byte[] result;
            if (data != null)
            {
                string jsonString = Newtonsoft.Json.JsonConvert.SerializeObject(data);
                byte[] jsonAsBytes = Encoding.UTF8.GetBytes(jsonString);
                byte[] lengthAsBytes = BitConverter.GetBytes(jsonAsBytes.Length);

                result = new byte[jsonAsBytes.Length + lengthAsBytes.Length];
                Array.Copy(lengthAsBytes, 0, result, 0, lengthAsBytes.Length);
                Array.Copy(jsonAsBytes, 0, result, lengthAsBytes.Length, jsonAsBytes.Length);
            }
            else
            {
                result = new byte[4];
            }
            stream.Write(result, 0, result.Length);
            stream.Flush();
        }

        bool parseBuffer()
        {
            if (receivedData.Count >= 4)
            {
                int msgLength = BitConverter.ToInt32(receivedData.GetRange(0, 4).ToArray(), 0);
                if (msgLength <= 0)
                {
                    receivedData.RemoveRange(0, 4);
                    if (!lastEmptyResponse)
                    {
                        OnDataReceived?.Invoke(new RemoteApiData());
                        lastEmptyResponse = true;
                    }
                    return true;
                }
                if (receivedData.Count >= msgLength + 4)
                {
                    byte[] msgAsBytes = receivedData.GetRange(4, msgLength).ToArray();
                    receivedData.RemoveRange(0, msgLength + 4);

                    string jsonMsg = Encoding.UTF8.GetString(msgAsBytes);
                    RemoteApiData inData = Newtonsoft.Json.JsonConvert.DeserializeObject<RemoteApiData>(jsonMsg);
                    if (inData != null)
                    {
                        OnDataReceived?.Invoke(inData);
                        lastEmptyResponse = false;
                        return true;
                    }
                }
            }
            return false;
        }
    }

    [Serializable]
    public class RemoteApiData
    {
        public string path;
        public float? duration;
        public float? currentTime;
        public float? playbackSpeed;
        public PlayerState? playerState;
    }

    public enum PlayerState
    {
        Play = 0,
        Pause = 1
    }
}